from dataclasses_json import DataClassJsonMixin
from dataclasses import dataclass
from typing import List, Optional


@dataclass
class Point(DataClassJsonMixin):
    def __post_init__(self) -> None:
        print(self, flush=True)

    x: float
    y: float


@dataclass
class CreateAoi(DataClassJsonMixin):
    def __post_init__(self) -> None:
        print(self, flush=True)

    name: str
    """Name of the AOI"""

    points: List[Point]
    z: float
    id: Optional[str] = None



# Standard library imports
import os

# Third party imports
import flask
import opengeode
from opengeodeweb_microservice.schemas import get_schemas_dict

# Local application imports
from opengeodeweb_back import geode_functions, utils_functions
from . import schemas
from opengeodeweb_back.geode_objects.geode_point_set3d import GeodePointSet3D
from opengeodeweb_back.geode_objects.geode_edged_curve3d import GeodeEdgedCurve3D

routes = flask.Blueprint("create", __name__, url_prefix="/create")
schemas_dict = get_schemas_dict(os.path.join(os.path.dirname(__file__), "schemas"))


@routes.route(
    schemas_dict["create_point"]["route"],
    methods=schemas_dict["create_point"]["methods"],
)
def create_point() -> flask.Response:
    """Endpoint to create a single point in 3D space."""
    json_data = utils_functions.validate_request(
        flask.request, schemas_dict["create_point"]
    )
    params = schemas.CreatePoint.from_dict(json_data)

    # Create the point
    pointset = GeodePointSet3D()
    builder = pointset.builder()
    builder.set_name(params.name)
    builder.create_point(opengeode.Point3D([params.x, params.y, params.z]))

    # Save and get info
    result = utils_functions.generate_native_viewable_and_light_viewable_from_object(
        pointset
    )
    return flask.make_response(result, 200)


@routes.route(
    schemas_dict["create_aoi"]["route"], methods=schemas_dict["create_aoi"]["methods"]
)
def create_aoi() -> flask.Response:
    """Endpoint to create an Area of Interest (AOI) as an EdgedCurve3D."""
    json_data = utils_functions.validate_request(
        flask.request, schemas_dict["create_aoi"]
    )
    params = schemas.CreateAoi.from_dict(json_data)

    # Create the edged curve
    edged_curve = GeodeEdgedCurve3D()
    builder = edged_curve.builder()
    builder.set_name(params.name)

    # Create vertices first
    for point in params.points:
        builder.create_point(opengeode.Point3D([point.x, point.y, params.z]))

    # Create edges between consecutive vertices and close the loop
    num_vertices = len(params.points)
    for i in range(num_vertices):
        next_i = (i + 1) % num_vertices
        builder.create_edge_with_vertices(i, next_i)

    # Save and get info
    result = utils_functions.generate_native_viewable_and_light_viewable_from_object(
        edged_curve
    )
    return flask.make_response(result, 200)


@routes.route(
    schemas_dict["create_voi"]["route"], methods=schemas_dict["create_voi"]["methods"]
)
def create_voi() -> flask.Response:
    """Endpoint to create a Volume of Interest (VOI) as an EdgedCurve3D (a bounding box/prism)."""
    json_data = utils_functions.validate_request(
        flask.request, schemas_dict["create_voi"]
    )
    params = schemas.CreateVoi.from_dict(json_data)

    aoi_data = geode_functions.get_data_info(params.aoi_id)
    if not aoi_data:
        flask.abort(404, f"AOI with id {params.aoi_id} not found")

    aoi_object = geode_functions.load_geode_object(params.aoi_id)
    if not isinstance(aoi_object, GeodeEdgedCurve3D):
        flask.abort(400, f"AOI with id {params.aoi_id} is not a GeodeEdgedCurve3D")

    aoi_curve = aoi_object.edged_curve
    nb_points = aoi_curve.nb_vertices()

    edged_curve = GeodeEdgedCurve3D()
    builder = edged_curve.builder()
    builder.set_name(params.name)

    for point_id in range(nb_points):
        aoi_point = aoi_curve.point(point_id)
        builder.create_point(
            opengeode.Point3D([aoi_point.value(0), aoi_point.value(1), params.z_min])
        )

    for point_id in range(nb_points):
        aoi_point = aoi_curve.point(point_id)
        builder.create_point(
            opengeode.Point3D([aoi_point.value(0), aoi_point.value(1), params.z_max])
        )

    for point_id in range(nb_points):
        next_point = (point_id + 1) % nb_points
        builder.create_edge_with_vertices(point_id, next_point)
        builder.create_edge_with_vertices(point_id + nb_points, next_point + nb_points)
        builder.create_edge_with_vertices(point_id, point_id + nb_points)

    result = utils_functions.generate_native_viewable_and_light_viewable_from_object(
        edged_curve
    )
    return flask.make_response(result, 200)



# Standard library imports
from __future__ import annotations
from abc import abstractmethod

# Third party imports
import opengeode as og

# Local application imports
from .types import ViewerType
from .geode_object import GeodeObject

ComponentRegistry = dict[og.ComponentType, list[og.uuid]]


class GeodeModel(GeodeObject):
    @classmethod
    def viewer_type(cls) -> ViewerType:
        return "model"

    @abstractmethod
    def mesh_components(self) -> ComponentRegistry: ...



from dataclasses_json import DataClassJsonMixin
from dataclasses import dataclass
from typing import Optional


@dataclass
class CreateVoi(DataClassJsonMixin):
    def __post_init__(self) -> None:
        print(self, flush=True)

    aoi_id: str
    """ID of the corresponding AOI"""

    name: str
    """Name of the VOI"""

    z_max: float
    """Maximum Z coordinate"""

    z_min: float
    """Minimum Z coordinate"""

    id: Optional[str] = None
    """Optional ID for updating existing VOI"""



# Standard library imports
from typing import Callable, Any

# Third party imports
from flask.testing import FlaskClient

# Local application imports

JsonData = dict[str, Any]


def test_route_wrong_params(
    client: FlaskClient, route: str, get_full_data: Callable[[], JsonData]
) -> None:
    for key, value in get_full_data().items():
        json = get_full_data()
        json.pop(key)
        response = client.post(route, json=json)
        assert response.status_code == 400
        error_description = response.get_json()["description"]
        assert "data must contain" in error_description
        assert f"'{key}'" in error_description

    json = get_full_data()
    json["dumb_key"] = "dumb_value"
    response = client.post(route, json=json)
    assert response.status_code == 400
    error_description = response.get_json()["description"]
    assert "data must not contain" in error_description
    assert "'dumb_key'" in error_description



import os
import shutil
import zipfile
import json
from flask.testing import FlaskClient
from werkzeug.datastructures import FileStorage
from pathlib import Path

from opengeodeweb_microservice.database.data import Data
from opengeodeweb_microservice.database.connection import get_session
from opengeodeweb_back import geode_functions
from opengeodeweb_back.geode_objects.geode_brep import GeodeBRep

base_dir = os.path.abspath(os.path.dirname(__file__))
data_dir = os.path.join(base_dir, "data")


def test_model_mesh_components(client: FlaskClient, test_id: str) -> None:
    route = "/opengeodeweb_back/models/vtm_component_indices"

    with client.application.app_context():
        data_path = geode_functions.data_file_path(test_id, "viewable.vtm")
        os.makedirs(os.path.dirname(data_path), exist_ok=True)
        shutil.copy(os.path.join(data_dir, "cube.vtm"), data_path)

    response = client.post(route, json={"id": test_id})
    assert response.status_code == 200

    uuid_dict = response.get_json()["uuid_to_flat_index"]
    assert isinstance(uuid_dict, dict)

    indices = list(uuid_dict.values())
    indices.sort()
    assert all(indices[i] > indices[i - 1] for i in range(1, len(indices)))
    for uuid in uuid_dict.keys():
        assert isinstance(uuid, str)


def test_extract_brep_uuids(client: FlaskClient, test_id: str) -> None:
    route = "/opengeodeweb_back/models/mesh_components"
    brep_filename = os.path.join(data_dir, "cube.og_brep")

    with client.application.app_context():
        data = Data.create(
            geode_object=GeodeBRep.geode_object_type(),
            viewer_object=GeodeBRep.viewer_type(),
            input_file=brep_filename,
        )
        data.native_file = brep_filename
        session = get_session()
        if session:
            session.commit()

        response = client.post(route, json={"id": data.id})
        assert response.status_code == 200
        assert "uuid_dict" in response.get_json()
        uuid_dict = response.get_json()["uuid_dict"]
        assert isinstance(uuid_dict, dict)


def test_export_project_route(client: FlaskClient, tmp_path: Path) -> None:
    route = "/opengeodeweb_back/export_project"
    snapshot = {
        "styles": {"1": {"visibility": True, "opacity": 1.0, "color": [0.2, 0.6, 0.9]}}
    }
    filename = "export_project_test.vease"
    project_folder = client.application.config["DATA_FOLDER_PATH"]
    os.makedirs(project_folder, exist_ok=True)
    database_root_path = os.path.join(project_folder, "project.db")
    with open(database_root_path, "wb") as f:
        f.write(b"test_project_db")

    with get_session() as session:
        session.query(Data).delete()
        session.commit()

        data1 = Data(
            id="test_data_1",
            geode_object="BRep",
            viewer_object="BRep",
            input_file="test_native.txt",
            native_file="test_native.txt",
            additional_files=[],
        )
        data2 = Data(
            id="test_data_2",
            geode_object="Section",
            viewer_object="Section",
            input_file="test_input.txt",
            native_file="test_native2.txt",
            additional_files=[],
        )
        session.add(data1)
        session.add(data2)
        session.commit()

        data1_dir = os.path.join(project_folder, "test_data_1")
        os.makedirs(data1_dir, exist_ok=True)
        with open(os.path.join(data1_dir, "test_native.txt"), "w") as f:
            f.write("native file content")

        data2_dir = os.path.join(project_folder, "test_data_2")
        os.makedirs(data2_dir, exist_ok=True)
        with open(os.path.join(data2_dir, "test_input.txt"), "w") as f:
            f.write("input file content")

    response = client.post(route, json={"snapshot": snapshot, "filename": filename})
    assert response.status_code == 200
    assert response.headers.get("new-file-name") == filename
    assert response.mimetype == "application/octet-binary"
    response.direct_passthrough = False
    zip_bytes = response.get_data()
    tmp_zip_path = tmp_path / filename
    tmp_zip_path.write_bytes(zip_bytes)

    with zipfile.ZipFile(tmp_zip_path, "r") as zip_file:
        names = zip_file.namelist()
        assert "snapshot.json" in names
        parsed = json.loads(zip_file.read("snapshot.json").decode("utf-8"))
        assert parsed == snapshot
        assert "project.db" in names
        assert "test_data_1/test_native.txt" in names
        assert "test_data_2/test_input.txt" in names

    response.close()

    export_path = os.path.join(project_folder, filename)
    if os.path.exists(export_path):
        os.remove(export_path)


def test_import_project_route(client: FlaskClient, tmp_path: Path) -> None:
    route = "/opengeodeweb_back/import_project"
    snapshot = {
        "styles": {"1": {"visibility": True, "opacity": 1.0, "color": [0.2, 0.6, 0.9]}}
    }

    original_data_folder = client.application.config["DATA_FOLDER_PATH"]
    client.application.config["DATA_FOLDER_PATH"] = os.path.join(
        str(tmp_path), "project_data"
    )
    db_path = os.path.join(client.application.config["DATA_FOLDER_PATH"], "project.db")

    import sqlite3, zipfile, json

    temp_db = tmp_path / "temp_project.db"
    conn = sqlite3.connect(str(temp_db))
    conn.execute(
        "CREATE TABLE datas (id TEXT PRIMARY KEY, geode_object TEXT, viewer_object TEXT, native_file TEXT, "
        "viewable_file TEXT, light_viewable_file TEXT, input_file TEXT, additional_files TEXT)"
    )
    conn.commit()
    conn.close()

    z = tmp_path / "import_project_test.vease"
    with zipfile.ZipFile(z, "w", compression=zipfile.ZIP_DEFLATED) as zipf:
        zipf.writestr("snapshot.json", json.dumps(snapshot))
        zipf.write(str(temp_db), "project.db")

    with open(z, "rb") as f:
        resp = client.post(
            route,
            data={"file": (f, "import_project_test.vease")},
            content_type="multipart/form-data",
        )

    assert resp.status_code == 200
    assert resp.get_json().get("snapshot") == snapshot
    assert os.path.exists(db_path)

    from opengeodeweb_microservice.database import connection

    client.application.config["DATA_FOLDER_PATH"] = original_data_folder
    test_db_path = os.environ.get("TEST_DB_PATH")
    if test_db_path:
        connection.init_database(test_db_path, create_tables=True)

    client.application.config["DATA_FOLDER_PATH"] = original_data_folder


def test_save_viewable_workflow_from_file(client: FlaskClient) -> None:
    file = os.path.join(data_dir, "cube.og_brep")
    upload_resp = client.put(
        "/opengeodeweb_back/upload_file",
        data={"file": FileStorage(open(file, "rb"))},
    )
    assert upload_resp.status_code == 201

    route = "/opengeodeweb_back/save_viewable_file"
    payload = {"geode_object_type": "BRep", "filename": "cube.og_brep"}

    response = client.post(route, json=payload)
    assert response.status_code == 200

    data_id = response.get_json()["id"]
    assert isinstance(data_id, str) and len(data_id) > 0
    assert response.get_json()["viewable_file"].endswith(".vtm")

    comp_resp = client.post(
        "/opengeodeweb_back/models/vtm_component_indices", json={"id": data_id}
    )
    assert comp_resp.status_code == 200

    refreshed = Data.get(data_id)
    assert refreshed is not None


def test_save_viewable_workflow_from_object(client: FlaskClient) -> None:
    route = "/opengeodeweb_back/create/create_aoi"
    aoi_data = {
        "name": "workflow_aoi",
        "points": [
            {"x": 0.0, "y": 0.0},
            {"x": 1.0, "y": 0.0},
            {"x": 1.0, "y": 1.0},
            {"x": 0.0, "y": 1.0},
        ],
        "z": 0.0,
    }

    response = client.post(route, json=aoi_data)
    assert response.status_code == 200

    data_id = response.get_json()["id"]
    assert isinstance(data_id, str) and len(data_id) > 0
    assert response.get_json()["geode_object_type"] == "EdgedCurve3D"
    assert response.get_json()["viewable_file"].endswith(".vtp")



# Standard library imports
import uuid

# Third party imports
import pytest
from flask.testing import FlaskClient

# Local application imports
from opengeodeweb_back import test_utils


@pytest.fixture
def point_data() -> test_utils.JsonData:
    return {"name": "test_point", "x": 1.0, "y": 2.0, "z": 3.0}


@pytest.fixture
def aoi_data() -> test_utils.JsonData:
    return {
        "name": "test_aoi",
        "points": [
            {"x": 0.0, "y": 0.0},
            {"x": 1.0, "y": 0.0},
            {"x": 1.0, "y": 1.0},
            {"x": 0.0, "y": 1.0},
        ],
        "z": 0.0,
    }


@pytest.fixture
def voi_data() -> test_utils.JsonData:
    """Fixture for Volume of Interest (VOI) test data."""
    return {
        "name": "test_voi",
        "aoi_id": str(uuid.uuid4()),
        "z_min": -50.0,
        "z_max": 100.0,
        "id": str(uuid.uuid4()),
    }


def test_create_point(client: FlaskClient, point_data: test_utils.JsonData) -> None:
    """Test the creation of a point with valid data."""
    route: str = "/opengeodeweb_back/create/create_point"

    # Test with all required data
    response = client.post(route, json=point_data)
    assert response.status_code == 200

    # Verify response data
    response_data = response.get_json()
    assert "viewable_file" in response_data
    assert "id" in response_data
    assert "name" in response_data
    assert "native_file" in response_data
    assert "viewer_type" in response_data
    assert "geode_object_type" in response_data

    assert response_data["name"] == point_data["name"]
    assert response_data["viewer_type"] == "mesh"
    assert response_data["geode_object_type"] == "PointSet3D"

    # Test with missing parameters
    test_utils.test_route_wrong_params(client, route, lambda: point_data.copy())


def test_create_aoi(client: FlaskClient, aoi_data: test_utils.JsonData) -> None:
    """Test the creation of an AOI with valid data."""
    route: str = "/opengeodeweb_back/create/create_aoi"

    # Test with all required data
    response = client.post(route, json=aoi_data)
    assert response.status_code == 200

    # Verify response data
    response_data = response.get_json()
    assert "viewable_file" in response_data
    assert "id" in response_data
    assert "name" in response_data
    assert "native_file" in response_data
    assert "viewer_type" in response_data
    assert "geode_object_type" in response_data

    assert response_data["name"] == aoi_data["name"]
    assert response_data["viewer_type"] == "mesh"
    assert response_data["geode_object_type"] == "EdgedCurve3D"

    # Test with missing parameters
    test_utils.test_route_wrong_params(client, route, lambda: aoi_data.copy())


def test_create_voi(
    client: FlaskClient, aoi_data: test_utils.JsonData, voi_data: test_utils.JsonData
) -> None:
    """Test the creation of a VOI with valid data (including optional id)."""
    aoi_route = "/opengeodeweb_back/create/create_aoi"
    aoi_response = client.post(aoi_route, json=aoi_data)
    assert aoi_response.status_code == 200
    aoi_id = aoi_response.get_json()["id"]

    voi_data["aoi_id"] = aoi_id

    voi_route = "/opengeodeweb_back/create/create_voi"
    response = client.post(voi_route, json=voi_data)
    assert response.status_code == 200

    response_data = response.get_json()
    assert "id" in response_data
    assert "name" in response_data
    assert response_data["name"] == voi_data["name"]
    assert response_data["viewer_type"] == "mesh"
    assert response_data["geode_object_type"] == "EdgedCurve3D"


def test_create_point_with_invalid_data(client: FlaskClient) -> None:
    """Test the point creation endpoint with invalid data."""
    route: str = "/opengeodeweb_back/create/create_point"

    # Test with non-numeric coordinates
    invalid_data: test_utils.JsonData = {
        "name": "invalid_point",
        "x": "not_a_number",
        "y": 2.0,
        "z": 3.0,
    }
    response = client.post(route, json=invalid_data)
    assert response.status_code == 400

    # Test with missing coordinates
    invalid_data = {"name": "invalid_point", "y": 2.0, "z": 3.0}
    response = client.post(route, json=invalid_data)
    assert response.status_code == 400


def test_create_aoi_with_invalid_data(
    client: FlaskClient, aoi_data: test_utils.JsonData
) -> None:
    """Test the AOI creation endpoint with invalid data."""
    route: str = "/opengeodeweb_back/create/create_aoi"

    invalid_data: test_utils.JsonData = {
        **aoi_data,
        "points": [
            {"x": "not_a_number", "y": 0.0},
            {"x": 1.0, "y": 0.0},
            {"x": 1.0, "y": 1.0},
            {"x": 0.0, "y": 1.0},
        ],
    }
    response = client.post(route, json=invalid_data)
    assert response.status_code == 400

    invalid_data = {**aoi_data, "points": [{"x": 0.0, "y": 0.0}, {"x": 1.0, "y": 0.0}]}
    response = client.post(route, json=invalid_data)
    assert response.status_code == 400

    invalid_data = {**aoi_data, "z": "not_a_number"}
    response = client.post(route, json=invalid_data)
    assert response.status_code == 400


def test_create_voi_with_invalid_data(
    client: FlaskClient, aoi_data: test_utils.JsonData, voi_data: test_utils.JsonData
) -> None:
    """Test the VOI creation endpoint with invalid data."""
    aoi_route = "/opengeodeweb_back/create/create_aoi"
    aoi_response = client.post(aoi_route, json=aoi_data)
    assert aoi_response.status_code == 200
    aoi_id = aoi_response.get_json()["id"]

    route = "/opengeodeweb_back/create/create_aoi"

    invalid_data = {**voi_data, "aoi_id": aoi_id, "z_min": "not_a_number"}
    response = client.post(route, json=invalid_data)
    assert response.status_code == 400

    invalid_data = {**voi_data, "aoi_id": aoi_id, "z_max": "not_a_number"}
    response = client.post(route, json=invalid_data)
    assert response.status_code == 400

    invalid_data = {**voi_data, "aoi_id": 12345}
    response = client.post(route, json=invalid_data)
    assert response.status_code == 400



import os
import xml.etree.ElementTree as ET
import flask
from opengeodeweb_microservice.schemas import get_schemas_dict

from opengeodeweb_back import geode_functions, utils_functions
from opengeodeweb_back.geode_objects.geode_model import GeodeModel
from . import schemas

routes = flask.Blueprint("models", __name__, url_prefix="/models")
schemas_dict = get_schemas_dict(os.path.join(os.path.dirname(__file__), "schemas"))


@routes.route(
    schemas_dict["vtm_component_indices"]["route"],
    methods=schemas_dict["vtm_component_indices"]["methods"],
)
def uuid_to_flat_index() -> flask.Response:
    json_data = utils_functions.validate_request(
        flask.request, schemas_dict["vtm_component_indices"]
    )
    params = schemas.VtmComponentIndices.from_dict(json_data)
    vtm_file_path = geode_functions.data_file_path(params.id, "viewable.vtm")
    tree = ET.parse(vtm_file_path)
    root = tree.find("vtkMultiBlockDataSet")
    if root is None:
        flask.abort(500, "Failed to read viewable file")
    uuid_to_flat_index = {}
    current_index = 0
    assert root is not None
    for elem in root.iter():
        if "uuid" in elem.attrib and elem.tag == "DataSet":
            uuid_to_flat_index[elem.attrib["uuid"]] = current_index
        current_index += 1
    return flask.make_response({"uuid_to_flat_index": uuid_to_flat_index}, 200)


@routes.route(
    schemas_dict["mesh_components"]["route"],
    methods=schemas_dict["mesh_components"]["methods"],
)
def extract_uuids_endpoint() -> flask.Response:
    json_data = utils_functions.validate_request(
        flask.request, schemas_dict["mesh_components"]
    )
    params = schemas.MeshComponents.from_dict(json_data)
    model = geode_functions.load_geode_object(params.id)
    if not isinstance(model, GeodeModel):
        flask.abort(400, f"{params.id} is not a GeodeModel")
    mesh_components = model.mesh_components()
    uuid_dict = {}
    for mesh_component, ids in mesh_components.items():
        component_name = mesh_component.get()
        uuid_dict[component_name] = [id.string() for id in ids]
    return flask.make_response({"uuid_dict": uuid_dict}, 200)
